"""
A namespace for operators used by this package's `Object` class.
"""

import builtins
import operator
import typing


class Operator:
    """Base class for enhanced operators."""
    def __init__(self, __f: typing.Callable, operation: str):
        self._f = __f
        self._operation = operation

    def __repr__(self):
        """Called for repr(self)."""
        return self._operation

    def __call__(self, *args, **kwds):
        """Called for self(*args, **kwds)."""
        return self._f(*args, **kwds)


eq = Operator(operator.eq, r'a == b')
ne = Operator(operator.ne, r'a != b')
lt = Operator(operator.lt, r'a < b')
le = Operator(operator.le, r'a <= b')
gt = Operator(operator.gt, r'a > b')
ge = Operator(operator.ge, r'a >= b')
abs = Operator(builtins.abs, r'abs(a)')
pos = Operator(operator.pos, r'+a')
neg = Operator(operator.neg, r'-a')
add = Operator(operator.add, r'a + b')
sub = Operator(operator.sub, r'a - b')
mul = Operator(operator.mul, r'a * b')
truediv = Operator(operator.truediv, r'a / b')
floordiv = Operator(operator.floordiv, r'a // b')
mod = Operator(operator.mod, r'a % b')
pow = Operator(builtins.pow, r'a ** b')




import abc
import numbers
import typing

import numpy.typing


@typing.runtime_checkable
class Real(typing.Protocol):
    """Abstract protocol for real-valued objects."""

    @abc.abstractmethod
    def __abs__(self):
        return NotImplemented

    @abc.abstractmethod
    def __pos__(self):
        return NotImplemented

    @abc.abstractmethod
    def __neg__(self):
        return NotImplemented

    @abc.abstractmethod
    def __eq__(self, other):
        return False

    @abc.abstractmethod
    def __ne__(self, other):
        return True

    @abc.abstractmethod
    def __le__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __lt__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __ge__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __gt__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __add__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __radd__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __sub__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __rsub__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __mul__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __rmul__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __truediv__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __rtruediv__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __floordiv__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __rfloordiv__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __mod__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __rmod__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __pow__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def __rpow__(self, other):
        return NotImplemented


DataType = typing.TypeVar(
    'DataType',
    int,
    float,
    numbers.Number,
    numpy.number,
    numpy.typing.ArrayLike,
    numpy.typing.NDArray,
)


class Object(Real, typing.Generic[DataType]):
    """A real-valued object with metadata attributes."""

    def __init__(
        self,
        __data: DataType,
        **metadata,
    ) -> None:
        if not isinstance(__data, Real):
            raise TypeError("Data input to Object must be real-valued")
        self._data = __data
        self._meta = metadata

    def __repr__(self):
        """Called for repr(self)."""
        try:
            datastr = numpy.array2string(
                self._data,
                separator=", ",
                threshold=6,
                edgeitems=2,
                prefix=f"{self.__class__.__qualname__}(",
                suffix=")"
            )
        except Exception:
            datastr = str(self._data)
        metastr = "metadata={" + ", ".join(f"{k!r}" for k in self._meta) + "}"
        return f"{self.__class__.__qualname__}({datastr}, {metastr})"

